/*
 * qbnrm.c
 *
 * Copyright (c) 2000,2001 Hajime Mochizuki <motizuki@jaist.ac.jp>
 * Natural Language Processing Laboratory,
 * Japan Advanced Institute of Science and Technology.
 * 
 * Use of this software is granted only for RESEARCH PURPOSES. 
 * Redistribution is granted as long as all copyright notices are retained.
 * Redistribution with modification is allowed provided that you make your 
 * modified version obviously distinguishable from the original one. 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES ARE DISCLAIMED.
 *
 */
#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <chainers/patrash.h>
#include <chainers/nero.h>
/* #include <chainers/cmn.h>*/
#include <chainers/mrpfil.h>

#include "tfflt.h"
#include "qbnrm.h"

int get_tf_from_hash (word, hashtab, size)
    char *word;
    struct hash_entry *hashtab[];
    int  size;
{
	unsigned int i;
	struct hash_entry *hp;

	i = hash (word, size);
	if (*(hashtab+i)!=NULL) {
		for (hp=*(hashtab+i); hp!=NULL; hp=hp->next) {
			if (strcmp(word, hp->word)==0) {
				return hp->tf;
			}
		}
	}
	return 0;
}

int chk_word_in_hash(word, htab, qhsize)
    char *word;
    struct hash_entry *htab[];
    int qhsize;
{
	unsigned int i;
	struct hash_entry *hp;

	i = hash (word, qhsize);
	if (*(htab+i)!=NULL) {
		for (hp=*(htab+i); hp!=NULL; hp=hp->next) {
			if (strcmp(word, hp->word)==0) {
				return 1;
			}
		}
	}
	return 0;
}

double norm_sscr (scr, wnum, mode, val, ave, std_dvi)
    double scr;
    int  wnum, mode;
    double val, ave, std_dvi;
{
	double newscr;

	if (scr==(double)0) return (double)0;

	switch (mode)
	{
	   case 0:	/* ʤ */
	default:
		newscr = scr;
		break;
	   case 1:	/* ññǳ */
		newscr = scr/(double)wnum;
		break;
	   case 2:	/* ʸкͤǳ */
		newscr = scr/(((double)10*
			       ((double)(wnum)-ave))/std_dvi+(double)50);
		break;
	   case 3:	/* ʸΥ2ñǳ */
		newscr = scr*scr/(double)wnum;
		break;
	   case 4:	/* ñ * norm_val ǳ */
		newscr = scr/((double)wnum * val);
		break;
	   case 5:	/* (ñǳ/2) + ʸкͤǳ */
		newscr = (scr/(double)wnum) + 
		     (scr/(((double)10*
			    ((double)(wnum)-ave))/std_dvi+(double)50));
		break;
	}
	return newscr;
}

void sscalcmore(ss, sn, awwct, mode, val)
     struct sentence *ss;
     int  awwct, mode;
     double val;
{
	int i, j, outct;
	double ave, std_dvi, ax;

	ax=(double)0;
	ave = (double)awwct/(double)sn, std_dvi=(double)0;

	if (mode==2||mode==5) {	/* кͤη׻ˤɸкɬ */
		for (i=1; i<=sn; i++) {
			ax+=((double)ss[i].wnum-ave)*((double)ss[i].wnum-ave);
		}
		std_dvi = sqrt(ax);
	}
	if (mode!=0) {	/* ʸĹˤ */
		for (i=1; i<=sn; i++) {
			ss[i].scr=norm_sscr(ss[i].scr, ss[i].wnum,
					    mode, val, ave, std_dvi);
		}
	}
	for (j=1; j<=sn; j++) {
		if (ss[j].sn==0) 
			printf ("%d %f\n", j, ss[j].scr);
		else 
			printf ("%d %f\n", ss[j].sn, ss[j].scr);
	}
}

void copy_words (tw, words, sn)
       struct tmpwords *tw;
       CHAFLT_OUT *words;
       int    sn;
{
	int i;

	if ((tw[sn].word = calloc((size_t)chaflt_out_num, sizeof(char **)))
	    ==NULL) {
		perror ("calloc for tw[sn].word");
		exit(1);
	}
	tw[sn].ct = chaflt_out_num;
	for (i=0; i<chaflt_out_num; i++) {
		tw[sn].word[i] = strdup(words[i].term);
	}
}
