/*
 * cfengine.c: 
 *
 * Copyright (c) 1998-2001 Hajime Mochizuki <motizuki@jaist.ac.jp>
 * Natural Language Processing Laboratory,
 * Japan Advanced Institute of Science and Technology.
 * 
 * Use of this software is granted only for RESEARCH PURPOSES. 
 * Redistribution is granted as long as all copyright notices are retained.
 * Redistribution with modification is allowed provided that you make your 
 * modified version obviously distinguishable from the original one. 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES ARE DISCLAIMED.
 *
 */
#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
/* #include "cmn.h"*/

#include <chainers/nero.h>
#include <chainers/mrpfil.h>
#include <chainers/chnfil.h>

#include "cfidfsw.h"
#include "tfflt.h"
#include "qbnrm.h"

extern struct patrash_files* patrash_preparation ();
extern struct patrash_answer_int* tell_patrash_int ();
/*extern int         */
int chaflt_out_num;
extern int         number_of_words;
extern CHAFLT_OUT* direct_chaflt ();
extern char*       chasen_sparse_tostr ();

extern void sscalcmore();
extern int get_tf_from_hash();
extern struct chn_entry* chn_sscanf();
extern void set_ppw_rules();
extern double get_postp_weight();

int    num_of_words;
int    num_of_sents;
int    chain_threshold;
int    gap_threshold;

struct snword *snw=NULL;

void snw_preparation (i)
     int i;
{
	if ((snw=calloc((size_t)(i+10), sizeof(*snw)))==NULL) {
		fprintf (stderr, "calloc for snw failed \n");
		exit(1);
	}
}

struct chain_data* cal_cdat (i)
     int i;
{
	struct chain_data *cdat;

	if ((cdat=calloc(i, sizeof(*cdat)))==NULL) {
		fprintf (stderr, "calloc for cdat failed\n");
		exit(1);
	}
	return cdat;
}

struct chain_data* ral_cdat (cdat, to)
     struct chain_data *cdat;     
     int to;
{
	struct chain_data *tmp;

	if ((tmp=realloc(cdat, sizeof(*cdat)*to))==NULL) {
		fprintf (stderr, "realloc for cdat failed\n");
		exit(1);
	}
	return tmp;
}

struct sentence* cal_and_fresh_ss (i)
     int i;
{
	int j;
	struct sentence *ss;

	if ((ss=calloc(i, sizeof(*ss)))==NULL) {
		fprintf (stderr, "calloc for ss failed\n");
		exit(1);
	}
	for (j=0; j<i; j++) {
		ss[i].scr=(double)0;
		ss[i].wnum=0;
	}
	return ss;
}
	
struct sentence* ral_and_fresh_ss (ss, from, to)
     struct sentence *ss;     
     int from, to;
{
	int i;
	struct sentence *tmp;

	if ((tmp=realloc(ss, sizeof(*ss)*to))==NULL) {
		fprintf (stderr, "realloc for ss failed\n");
		exit(1);
	}
	ss=tmp;
	for (i=from; i<to; i++) {
		ss[i].scr=(double)0;
		ss[i].wnum=0;
	}
	return ss;
}

/*
 * Ϣξơsnw롥
 */
struct snword* fill_snw (buf, snw, tf, chn_type, no_pw)
     char *buf;
     struct snword *snw;
     int tf, chn_type, no_pw;
{
	int i;
	struct chn_entry *ce;

	ce=NULL;
	ce = chn_sscanf(buf, chn_type);
	(snw+tf)->sn=ce->sn;

	switch (chn_type) {
	case CHNTYPE_COOCCURRENCE:
		(snw+tf)->word= strdup(ce->field[1]);
		(snw+tf)->sense= strdup(ce->field[1]);
		break;
	case CHNTYPE_THESAURUS:
		(snw+tf)->word= strdup(ce->field[1]);
		(snw+tf)->sense= strdup(ce->field[0]);
		break;
	}
	if (no_pw) (snw+tf)->ppw = (double) 1.0;
	else       (snw+tf)->ppw = get_postp_weight (ce->field[2], ce->field[3]);

	free(ce); ce=NULL;
	return snw;
}

struct sentence* fill_ss (ss, snw, tf)
     struct sentence *ss;
     struct snword *snw;
     int tf;
{
	int sn;

	sn=(snw+tf)->sn;
	(ss+sn)->sn=sn;
	(ss+sn)->wnum = ((ss+sn)->wnum) + 1;
	return ss;
}

double calc_cscr (tf, snw, cdat, cn, N, df, qper, cw_m)
    int tf, cn, N, df, cw_m; /* cw_m=chnw_mode */
    struct snword *snw;
    struct chain_data *cdat;
    double qper;
{
	int i;
	double scr;
	struct snword *tsnw;
	
	if (cw_m==0)      scr = (double)1*qper;
	else if (cw_m==1) scr = (double)tf*qper;
	else if (cw_m==2) scr = (double)tf * log((double)N/(double)df) *qper;
	else if (cw_m==3) scr = qper *(double)tf * log((double)N/(double)df);

	if ((tsnw=calloc((size_t)tf, sizeof(struct snword)))==NULL) {
		perror ("calloc for tsnw");
		exit(1);
	}
	for (i=0; i<tf; i++) {
		tsnw[i].sn   = snw[i].sn;
		tsnw[i].word = strdup(snw[i].word);
		tsnw[i].sense= strdup(snw[i].sense);
		/*strcpy(tsnw[i].pp, snw[i].pp);*/
		tsnw[i].ppw = snw[i].ppw;
	}
	cdat[cn].cwn = tf;
	cdat[cn].score = scr;
	cdat[cn].snw = tsnw;

	return scr;
}

/*
 * NȤDF֤ͤ
 */
int get_num (word, pf)
     char *word;
     struct patrash_files *pf;
{
        struct patrash_answer_int *intr;
	int num;
		
	if ((intr = tell_patrash_int(word, pf->pa, pf->wan))==NULL) {
		fprintf (stderr, "??? dffile is broken??\n");
		exit(1);
	}
	num = intr->num; free(intr); intr=NULL;
	return num;
}


/*
 * query biasedѤqueryեɤ߹ǥϥå;
 */
int query_preparation(fn, qhtab, qhsz)
     char *fn;
     struct hash_entry *qhtab;
     int qhsz;
{
	int a;
	FILE *f;
	char buf[BUFSIZ], QWD[100], QTF[20];

	if ((f=fopen(fn, "r"))==NULL) {
		fprintf (stderr, "cannot open %s\n", fn);
		exit(1);
	}
	init_hash_f(qhtab, qhsz);
	a=0;
	while (fgets(buf, BUFSIZ, f)!=NULL) {
		sscanf(buf, "%s %s\n", QWD, QTF);
		a+=hash_update(QWD, QTF, qhtab, qhsz);
	}
	fclose(f);
	return a;
}


void calc_sscr (cdat, cn, ss, acscr)
    struct chain_data *cdat;
    int cn;
    struct sentence *ss;
    double acscr;

{
	int i, j, ppi;
	double ave, thre;

	ave = acscr / (double)cn;

	if (thre_mode==0) {
		thre = (double)0;
	} else if (thre_mode==1) {
		thre = ave*thre_val;
		fprintf (stderr, "acscr =%f cn = %d ave = %f thre = %f\n", acscr, cn, ave, thre); 
	} else {
		thre = thre_val;
	}
	for (i=0; i<cn; i++) {
		if (cdat[i].score>=thre) {	/* Ͱʾ */
			for (j=0; j<cdat[i].cwn; j++) {
				ss[cdat[i].snw[j].sn].scr+=
					cdat[i].score * cdat[i].snw[j].ppw;
			}
		}
	}
}

void main_engine (qhtab, qhsz, qentry, mrf, dffile)
     struct hash_entry *qhtab;
     int qhsz, qentry;
     char *mrf, *dffile;
{
	int sn, wn, wnum, cdct, ssct, cn, N, df, df0, tf, awwct;
	double acscr, qper;
	struct chain_data *cdat;
	struct sentence *ss;
	struct chn_entry *ce;
        struct patrash_files *pf;
        struct patrash_answer_int *intr;
	char buf[BUFSIZ];

        fgets(buf, BUFSIZ, stdin); /* ñ ʸ ɤ߹ */
	sscanf(buf, "%d %d %d\n", &wn, &sn, &wnum);
	snw_preparation(wn);
	cdct = 30;
	cdat = cal_cdat(cdct);
	ssct = sn+1;
	ss = cal_and_fresh_ss(ssct);
	cn=0, N=0, tf=0, df=1, acscr=(double)0;

	if (chnw_mode==2) {
		pf = patrash_preparation(dffile);
		N  = get_num("N", pf);
	}
	awwct=0; qper=(double)1;
	set_ppw_rules (mrp_type, mrf);
	while(fgets(buf, BUFSIZ, stdin)!=NULL) {
		if (buf[0]=='\n') { /* ϢѤ */
			if (chnw_mode==2 && chn_type==0) {
				df = get_num(snw[0].sense, pf);
			}
			if (tf>=min_chntf) {
				if (cn>=cdct) {
					cdat = ral_cdat(cdat, cdct+10);
					cdct+=10;
				}
				acscr+=calc_cscr(tf,snw,cdat,cn,N,df,qper,chnw_mode);
				cn++;
			}
			tf=0, df=1, qper=(double)1;
		} else {
			snw = fill_snw(buf, snw, tf, chn_type, no_pw);
			ss  = fill_ss(ss, snw, tf);
			awwct+=ss[snw[tf].sn].wnum;
			if (qb)	if (chk_word_in_hash(snw[tf].word,qhtab,qhsz)) qper=qwgt;

			if (chnw_mode==2 && chn_type!=0) { /* dfĤ */
				df0 = get_num(snw[tf].sense, pf);
				if (df < df0) df=df0;
			}
			tf++;
		}
	}
	calc_sscr (cdat,cn,ss,acscr);
	sscalcmore(ss, sn, awwct,norm_mode,norm_val);
}
